// Copyright 2023 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "content/browser/interest_group/bidding_and_auction_server_key_fetcher.h"

#include "base/files/scoped_temp_dir.h"
#include "base/functional/callback_helpers.h"
#include "base/memory/scoped_refptr.h"
#include "base/run_loop.h"
#include "base/strings/stringprintf.h"
#include "base/test/bind.h"
#include "base/test/scoped_feature_list.h"
#include "base/test/task_environment.h"
#include "base/time/time.h"
#include "content/browser/interest_group/interest_group_features.h"
#include "content/browser/interest_group/interest_group_manager_impl.h"
#include "services/data_decoder/public/cpp/test_support/in_process_data_decoder.h"
#include "services/network/public/cpp/shared_url_loader_factory.h"
#include "services/network/public/cpp/weak_wrapper_shared_url_loader_factory.h"
#include "services/network/test/test_url_loader_factory.h"
#include "testing/gmock/include/gmock/gmock.h"
#include "testing/gtest/include/gtest/gtest.h"
#include "third_party/blink/public/common/features.h"

namespace content {

namespace {

const char kDefaultGCPKeyURL[] = "https://example.com/default_keys";
const char kOtherDefaultGCPKeyURL[] = "https://example.com/other_keys";

const char kCoordinator1[] = "https://first.coordinator.test/";
const char kCoordinator1KeyURL[] = "https://example.com/first_keys";

const char kCoordinator2[] = "https://second.coordinator.test/";
const char kCoordinator2KeyURL[] = "https://example.com/second_keys";

class BiddingAndAuctionServerKeyFetcherTest : public testing::Test {
 public:
  void SetUp() override {
    std::string kKeyConfig = base::StringPrintf(
        R"({
"%s": "%s",
"%s": "%s",
"%s": "%s"
      })",
        kCoordinator1, kCoordinator1KeyURL, kCoordinator2, kCoordinator2KeyURL,
        kDefaultBiddingAndAuctionGCPCoordinatorOrigin, kDefaultGCPKeyURL);
    feature_list_.InitWithFeaturesAndParameters(
        /*enabled_features=*/{{blink::features::kFledgeBiddingAndAuctionServer,
                               {{"FledgeBiddingAndAuctionKeyConfig",
                                 kKeyConfig}}}},
        /*disabled_features=*/{});
    ASSERT_TRUE(temp_directory_.CreateUniqueTempDir());
    manager_ = std::make_unique<InterestGroupManagerImpl>(
        temp_directory_.GetPath(), false,
        InterestGroupManagerImpl::ProcessMode::kDedicated, nullptr,
        base::NullCallback());
    shared_url_loader_factory_ =
        base::MakeRefCounted<network::WeakWrapperSharedURLLoaderFactory>(
            &url_loader_factory_);
  }

  url::Origin CoordinatorOrigin() {
    return url::Origin::Create(
        GURL(kDefaultBiddingAndAuctionGCPCoordinatorOrigin));
  }

 protected:
  network::TestURLLoaderFactory url_loader_factory_;
  scoped_refptr<network::SharedURLLoaderFactory> shared_url_loader_factory_;
  base::test::TaskEnvironment task_environment_{
      base::test::TaskEnvironment::TimeSource::MOCK_TIME};
  base::test::ScopedFeatureList feature_list_;
  data_decoder::test::InProcessDataDecoder decoder_;
  base::ScopedTempDir temp_directory_;
  std::unique_ptr<InterestGroupManagerImpl> manager_;

  std::pair<base::Time, std::vector<BiddingAndAuctionServerKey>>
  GetDBStoredKeysWithExpiration(const url::Origin& coordinator) {
    std::pair<base::Time, std::vector<BiddingAndAuctionServerKey>>
        expiration_and_keys;
    base::RunLoop run_loop;
    manager_->GetBiddingAndAuctionServerKeys(
        coordinator,
        base::BindLambdaForTesting(
            [&](std::pair<base::Time, std::vector<BiddingAndAuctionServerKey>>
                    stored_keys) {
              expiration_and_keys = std::move(stored_keys);
              run_loop.Quit();
            }));
    run_loop.Run();
    return expiration_and_keys;
  }

  void SetDBStoredKeys(const url::Origin& coordinator,
                       std::vector<BiddingAndAuctionServerKey> keys,
                       base::Time expiration) {
    manager_->SetBiddingAndAuctionServerKeys(coordinator, keys, expiration);
  }

  content::BiddingAndAuctionServerKeyFetcher CreateFetcher() {
    return BiddingAndAuctionServerKeyFetcher(manager_.get());
  }
};

TEST_F(BiddingAndAuctionServerKeyFetcherTest, UnknownCoordinator) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitWithFeaturesAndParameters(
      /*enabled_features=*/{{blink::features::kFledgeBiddingAndAuctionServer,
                             {{"FledgeBiddingAndAuctionKeyConfig", ""},
                              {"FledgeBiddingAndAuctionKeyURL", ""}}}},
      /*disabled_features=*/{});
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, url::Origin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_FALSE(maybe_key.has_value());
        run_loop.Quit();
      }));
  run_loop.Run();
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, NoURL) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitWithFeaturesAndParameters(
      /*enabled_features=*/{{blink::features::kFledgeBiddingAndAuctionServer,
                             {{"FledgeBiddingAndAuctionKeyConfig", ""},
                              {"FledgeBiddingAndAuctionKeyURL", ""}}}},
      /*disabled_features=*/{});
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_FALSE(maybe_key.has_value());
        run_loop.Quit();
      }));
  run_loop.Run();
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, BadResponses) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  std::vector<std::string> bad_responses = {
      "",
      "abc",  // Not valid JSON
      "1",
      R"("keys")",
      "[]",
      "{}",
      R"({ "keys": {}})",
      R"({ "keys": []})",
      R"({ "keys": [{ "key": 1}]})",
      R"({ "keys": [{ "key": "abc/_#$\n\n"}]})",
      R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA="}]})",
      R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
      "id": 1
      }]})",
      R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
      "id": ""
      }]})"};

  for (const auto& response : bad_responses) {
    SCOPED_TRACE(response);
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_FALSE(maybe_key.has_value());
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(url_loader_factory_.SimulateResponseForPendingRequest(
        kDefaultGCPKeyURL, response));
    run_loop.Run();
  }
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, FailsAll) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
  content::BiddingAndAuctionServerKey key1, key2;
  int completed = 0;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_FALSE(maybe_key.has_value());
        completed++;
      }));
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_FALSE(maybe_key.has_value());
        completed++;
        run_loop.Quit();
      }));
  EXPECT_TRUE(url_loader_factory_.SimulateResponseForPendingRequest(
      kDefaultGCPKeyURL, "", net::HTTP_NOT_FOUND));
  run_loop.Run();
  EXPECT_EQ(2, completed);
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, RequestDuringFailure) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
  content::BiddingAndAuctionServerKey key1, key2;
  int completed = 0;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_FALSE(maybe_key.has_value());
        completed++;
        fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                              base::BindLambdaForTesting(
                                  [&](base::expected<BiddingAndAuctionServerKey,
                                                     std::string> maybe_key) {
                                    EXPECT_FALSE(maybe_key.has_value());
                                    completed++;
                                    run_loop.Quit();
                                  }));
      }));
  EXPECT_TRUE(url_loader_factory_.SimulateResponseForPendingRequest(
      kDefaultGCPKeyURL, "", net::HTTP_NOT_FOUND));
  run_loop.Run();
  EXPECT_EQ(2, completed);
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, GoodResponse) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  content::BiddingAndAuctionServerKey key;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_TRUE(maybe_key.has_value());
        key = *maybe_key;
        run_loop.Quit();
      }));
  EXPECT_TRUE(
      url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                            R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
      "id": "12345678-9abc-def0-1234-56789abcdef0"
      }]})"));
  run_loop.Run();
  EXPECT_EQ(0x12, key.id);
  EXPECT_EQ(std::string(32, '\0'), key.key);
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, RequestDuringSuccess) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
  int completed = 0;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_TRUE(maybe_key.has_value());
        completed++;
        fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                              base::BindLambdaForTesting(
                                  [&](base::expected<BiddingAndAuctionServerKey,
                                                     std::string> maybe_key) {
                                    EXPECT_TRUE(maybe_key.has_value());
                                    completed++;
                                    run_loop.Quit();
                                  }));
      }));
  EXPECT_TRUE(
      url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                            R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
      "id": "12345678-9abc-def0-1234-56789abcdef0"
      }]})"));
  run_loop.Run();
  EXPECT_EQ(2, completed);
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, CachesValue) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key.id);
    EXPECT_EQ(std::string(32, '\0'), key.key);
  }

  {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    // Shouldn't use this response (it should still be cached).
    EXPECT_FALSE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAE\u003d",
        "id": "23456789-abcd-ef01-2345-6789abcdef01"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key.id);
    EXPECT_EQ(std::string(32, '\0'), key.key);
  }
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, ReadsValuesCachedInDBIfEnabled) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitAndEnableFeature(features::kFledgeStoreBandAKeysInDB);

  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
  std::vector<BiddingAndAuctionServerKey> keys;
  BiddingAndAuctionServerKey key;
  key.id = 1;
  key.key = "a";
  keys.push_back(key);
  SetDBStoredKeys(CoordinatorOrigin(), keys, base::Time::Now() + base::Days(2));
  task_environment_.RunUntilIdle();

  {
    content::BiddingAndAuctionServerKey returned_key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                returned_key = *maybe_key;
                                run_loop.Quit();
                              }));
    // Shouldn't use this response (it should have been retrieved from the
    // database).
    EXPECT_FALSE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAE\u003d",
        "id": "23456789-abcd-ef01-2345-6789abcdef01"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(key.id, returned_key.id);
    EXPECT_EQ(key.key, returned_key.key);
  }

  // Fast forward past the expiration time stored in the database to make sure
  // it was respected.
  task_environment_.FastForwardBy(base::Days(2) + base::Seconds(1));

  {
    // This should make a new request to the network now.
    content::BiddingAndAuctionServerKey returned_key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                returned_key = *maybe_key;
                                run_loop.Quit();
                              }));
    task_environment_.RunUntilIdle();
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAE\u003d",
        "id": "23456789-abcd-ef01-2345-6789abcdef01"
        }]})"));
    run_loop.Run();
  }
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, WritesValuesToDBIfEnabled) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitAndEnableFeature(features::kFledgeStoreBandAKeysInDB);

  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  content::BiddingAndAuctionServerKey key;
  {
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    // Make sure the task environment runs past the database fetch.
    while (!url_loader_factory_.IsPending(kDefaultGCPKeyURL)) {
      // IsPending calls RunUntilIdle().
    }
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key.id);
    EXPECT_EQ(std::string(32, '\0'), key.key);
  }
  task_environment_.RunUntilIdle();
  std::pair<base::Time, std::vector<content::BiddingAndAuctionServerKey>>
      expiration_and_stored_keys =
          GetDBStoredKeysWithExpiration(CoordinatorOrigin());
  EXPECT_EQ(base::Time::Now() + base::Days(7),
            expiration_and_stored_keys.first);
  ASSERT_EQ(1u, expiration_and_stored_keys.second.size());
  EXPECT_EQ(expiration_and_stored_keys.second[0].key, key.key);
  EXPECT_EQ(expiration_and_stored_keys.second[0].id, key.id);
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest,
       MaybePrefetchKeysFailureFailsPendingGetOrFetchKey) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitAndEnableFeature(features::kFledgePrefetchBandAKeys);

  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  fetcher.MaybePrefetchKeys(shared_url_loader_factory_);

  {
    bool completed = false;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_FALSE(maybe_key.has_value());
                                completed = true;
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(url_loader_factory_.SimulateResponseForPendingRequest(
        kDefaultGCPKeyURL, "", net::HTTP_NOT_FOUND));
    run_loop.Run();
    EXPECT_TRUE(completed);
  }

  // The following GetOrFetchKey after the prefetch failure should be
  // successful.
  {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key.id);
    EXPECT_EQ(std::string(32, '\0'), key.key);
  }
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, MaybePrefetchKeysCachesValue) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitAndEnableFeature(features::kFledgePrefetchBandAKeys);

  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  fetcher.MaybePrefetchKeys(shared_url_loader_factory_);
  task_environment_.RunUntilIdle();
  // MaybePrefetchKeys will try to fetch all keys in the
  // FledgeBiddingAndAuctionKeyConfig.
  EXPECT_TRUE(url_loader_factory_.IsPending(kDefaultGCPKeyURL));
  EXPECT_TRUE(url_loader_factory_.IsPending(kCoordinator1KeyURL));
  EXPECT_TRUE(url_loader_factory_.IsPending(kCoordinator2KeyURL));
  EXPECT_TRUE(
      url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                            R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
  task_environment_.RunUntilIdle();

  {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    // Shouldn't use this response (it should still be cached).
    EXPECT_FALSE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAE\u003d",
        "id": "23456789-abcd-ef01-2345-6789abcdef01"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key.id);
    EXPECT_EQ(std::string(32, '\0'), key.key);
  }

  // Keys should not be fetched a second time.
  fetcher.MaybePrefetchKeys(shared_url_loader_factory_);
  task_environment_.RunUntilIdle();
  EXPECT_FALSE(url_loader_factory_.IsPending(kDefaultGCPKeyURL));
  EXPECT_FALSE(
      url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                            R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest,
       MaybePrefetchKeysUpdatesExpiredValue) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitAndEnableFeature(features::kFledgePrefetchBandAKeys);

  // Get a key that we will let expire.
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
  {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key.id);
    EXPECT_EQ(std::string(32, '\0'), key.key);
  }

  // Let the key expire.
  task_environment_.RunUntilIdle();
  task_environment_.FastForwardBy(base::Days(7) + base::Seconds(1));

  fetcher.MaybePrefetchKeys(shared_url_loader_factory_);
  task_environment_.RunUntilIdle();
  // We should still try fetching all the keys, including for kDefaultGCPKeyURL.
  EXPECT_TRUE(url_loader_factory_.IsPending(kDefaultGCPKeyURL));
  EXPECT_TRUE(url_loader_factory_.IsPending(kCoordinator1KeyURL));
  EXPECT_TRUE(url_loader_factory_.IsPending(kCoordinator2KeyURL));
  EXPECT_TRUE(
      url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                            R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAE\u003d",
        "id": "23456789-abcd-ef01-2345-6789abcdef01"
        }]})"));
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest,
       MaybePrefetchKeysDoesNotCacheValueIfFeatureDisabled) {
  {
    // Disable kFledgePrefetchBandAKeys.
    base::test::ScopedFeatureList feature_list;
    feature_list.InitAndDisableFeature(features::kFledgePrefetchBandAKeys);

    content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
    fetcher.MaybePrefetchKeys(shared_url_loader_factory_);
    task_environment_.RunUntilIdle();
    EXPECT_FALSE(url_loader_factory_.IsPending(kDefaultGCPKeyURL));
    EXPECT_FALSE(url_loader_factory_.IsPending(kCoordinator1KeyURL));
    EXPECT_FALSE(url_loader_factory_.IsPending(kCoordinator2KeyURL));
  }

  {
    // Disable kFledgeBiddingAndAuctionServer.
    base::test::ScopedFeatureList feature_list;

    feature_list.InitWithFeatures(
        {features::kFledgePrefetchBandAKeys},
        {blink::features::kFledgeBiddingAndAuctionServer});
    content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();
    fetcher.MaybePrefetchKeys(shared_url_loader_factory_);
    task_environment_.RunUntilIdle();
    EXPECT_FALSE(url_loader_factory_.IsPending(kDefaultGCPKeyURL));
    EXPECT_FALSE(url_loader_factory_.IsPending(kCoordinator1KeyURL));
    EXPECT_FALSE(url_loader_factory_.IsPending(kCoordinator2KeyURL));
  }
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, CoalescesRequests) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  {
    content::BiddingAndAuctionServerKey key1, key2;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key1 = *maybe_key;
                              }));
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                key2 = *maybe_key;
                                EXPECT_TRUE(maybe_key.has_value());
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }]})"));
    run_loop.Run();
    EXPECT_EQ(0x12, key1.id);
    EXPECT_EQ(std::string(32, '\0'), key1.key);

    EXPECT_EQ(0x12, key2.id);
    EXPECT_EQ(std::string(32, '\0'), key2.key);
  }
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, ChoosesRandomKey) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    EXPECT_TRUE(
        url_loader_factory_.SimulateResponseForPendingRequest(kDefaultGCPKeyURL,
                                                              R"({ "keys": [{
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
        "id": "12345678-9abc-def0-1234-56789abcdef0"
        }, {
        "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAE=",
        "id": "23456789-abcd-ef01-2345-6789abcdef01"
        }]})"));
    run_loop.Run();
  }

  std::set<uint8_t> ids;
  while (ids.size() < 2) {
    content::BiddingAndAuctionServerKey key;
    base::RunLoop run_loop;
    fetcher.GetOrFetchKey(shared_url_loader_factory_, CoordinatorOrigin(),
                          base::BindLambdaForTesting(
                              [&](base::expected<BiddingAndAuctionServerKey,
                                                 std::string> maybe_key) {
                                EXPECT_TRUE(maybe_key.has_value());
                                key = *maybe_key;
                                run_loop.Quit();
                              }));
    run_loop.Run();
    ids.insert(key.id);
  }
  EXPECT_THAT(ids, testing::ElementsAre(0x12, 0x23));
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, OverridesConfig) {
  base::test::ScopedFeatureList feature_list;
  std::string kKeyConfig = base::StringPrintf(
      R"({
"%s": "%s",
"%s": "%s",
"%s": "%s"
      })",
      kCoordinator1, kCoordinator1KeyURL, kCoordinator2, kCoordinator2KeyURL,
      kDefaultBiddingAndAuctionGCPCoordinatorOrigin, kDefaultGCPKeyURL);
  feature_list.InitWithFeaturesAndParameters(
      /*enabled_features=*/{{blink::features::kFledgeBiddingAndAuctionServer,
                             {{"FledgeBiddingAndAuctionKeyConfig", kKeyConfig},
                              {"FledgeBiddingAndAuctionKeyURL",
                               kOtherDefaultGCPKeyURL}}}},
      /*disabled_features=*/{});
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  content::BiddingAndAuctionServerKey key;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_TRUE(maybe_key.has_value());
        key = *maybe_key;
        run_loop.Quit();
      }));
  EXPECT_TRUE(url_loader_factory_.SimulateResponseForPendingRequest(
      kOtherDefaultGCPKeyURL,
      R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
      "id": "12345678-9abc-def0-1234-56789abcdef0"
      }]})"));
  run_loop.Run();
  EXPECT_EQ(0x12, key.id);
  EXPECT_EQ(std::string(32, '\0'), key.key);
}

TEST_F(BiddingAndAuctionServerKeyFetcherTest, NoConfigOnlyURL) {
  base::test::ScopedFeatureList feature_list;
  feature_list.InitWithFeaturesAndParameters(
      /*enabled_features=*/{{blink::features::kFledgeBiddingAndAuctionServer,
                             {{"FledgeBiddingAndAuctionKeyURL",
                               kOtherDefaultGCPKeyURL}}}},
      /*disabled_features=*/{});
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  content::BiddingAndAuctionServerKey key;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, CoordinatorOrigin(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_TRUE(maybe_key.has_value());
        key = *maybe_key;
        run_loop.Quit();
      }));
  EXPECT_TRUE(url_loader_factory_.SimulateResponseForPendingRequest(
      kOtherDefaultGCPKeyURL,
      R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
      "id": "12345678-9abc-def0-1234-56789abcdef0"
      }]})"));
  run_loop.Run();
  EXPECT_EQ(0x12, key.id);
  EXPECT_EQ(std::string(32, '\0'), key.key);
}

class BiddingAndAuctionServerKeyFetcherCoordinatorTest
    : public BiddingAndAuctionServerKeyFetcherTest,
      public ::testing::WithParamInterface<int> {
 public:
  url::Origin GetCoordinator() {
    switch (GetParam()) {
      case 0:
        return url::Origin::Create(
            GURL(kDefaultBiddingAndAuctionGCPCoordinatorOrigin));
      case 1:
        return url::Origin::Create(GURL(kCoordinator1));
      case 2:
        return url::Origin::Create(GURL(kCoordinator2));
    }
    NOTREACHED_NORETURN();
  }

  std::string GetURL() {
    switch (GetParam()) {
      case 0:
        return kDefaultGCPKeyURL;
      case 1:
        return kCoordinator1KeyURL;
      case 2:
        return kCoordinator2KeyURL;
    }
    NOTREACHED_NORETURN();
  }
};

TEST_P(BiddingAndAuctionServerKeyFetcherCoordinatorTest, GoodResponse) {
  content::BiddingAndAuctionServerKeyFetcher fetcher = CreateFetcher();

  content::BiddingAndAuctionServerKey key;
  base::RunLoop run_loop;
  fetcher.GetOrFetchKey(
      shared_url_loader_factory_, GetCoordinator(),
      base::BindLambdaForTesting([&](base::expected<BiddingAndAuctionServerKey,
                                                    std::string> maybe_key) {
        EXPECT_TRUE(maybe_key.has_value());
        key = *maybe_key;
        run_loop.Quit();
      }));
  EXPECT_TRUE(
      url_loader_factory_.SimulateResponseForPendingRequest(GetURL(),
                                                            R"({ "keys": [{
      "key": "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\u003d",
      "id": "12345678-9abc-def0-1234-56789abcdef0"
      }]})"));
  run_loop.Run();
  EXPECT_EQ(0x12, key.id);
  EXPECT_EQ(std::string(32, '\0'), key.key);
}

INSTANTIATE_TEST_SUITE_P(
    /* no label */,
    BiddingAndAuctionServerKeyFetcherCoordinatorTest,
    ::testing::Range(0, 3));

}  // namespace
}  // namespace content
